@@ -1,14 +1,15 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #define _POSIX_C_SOURCE 200809L
+#include "menu/menu.h"
 #include <assert.h>
 #include <libxml/parser.h>
-#include <libxml/tree.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
-#include <wayland-server-core.h>
+#include <unistd.h>
+#include <wlr/types/wlr_scene.h>
 #include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/log.h>
 #include "action.h"
@@ -17,33 +18,26 @@
 #include "common/font.h"
 #include "common/lab-scene-rect.h"
 #include "common/list.h"
-#include "common/macros.h"
 #include "common/mem.h"
-#include "common/nodename.h"
-#include "common/scaled-font-buffer.h"
-#include "common/scaled-icon-buffer.h"
-#include "common/scene-helpers.h"
 #include "common/spawn.h"
 #include "common/string-helpers.h"
+#include "common/xml.h"
+#include "config/rcxml.h"
 #include "labwc.h"
-#include "menu/menu.h"
-#include "workspaces.h"
-#include "view.h"
 #include "node.h"
+#include "output.h"
+#include "scaled-buffer/scaled-font-buffer.h"
+#include "scaled-buffer/scaled-icon-buffer.h"
 #include "theme.h"
+#include "translate.h"
+#include "view.h"
+#include "workspaces.h"
 
 #define PIPEMENU_MAX_BUF_SIZE 1048576  /* 1 MiB */
 #define PIPEMENU_TIMEOUT_IN_MS 4000    /* 4 seconds */
 
 #define ICON_SIZE (rc.theme->menu_item_height - 2 * rc.theme->menu_items_padding_y)
 
-/* state-machine variables for processing <item></item> */
-static bool in_item;
-static struct menuitem *current_item;
-static struct action *current_item_action;
-
-static struct menu *current_menu;
-
 static bool waiting_for_pipe_menu;
 static struct menuitem *selected_item;
 
@@ -72,7 +66,8 @@ is_unique_id(struct server *server, const char *id)
 }
 
 static struct menu *
-menu_create(struct server *server, const char *id, const char *label)
+menu_create(struct server *server, struct menu *parent, const char *id,
+		const char *label)
 {
 	if (!is_unique_id(server, id)) {
 		wlr_log(WLR_ERROR, "menu id %s already exists", id);
@@ -84,7 +79,7 @@ menu_create(struct server *server, const char *id, const char *label)
 	wl_list_init(&menu->menuitems);
 	menu->id = xstrdup(id);
 	menu->label = xstrdup(label ? label : id);
-	menu->parent = current_menu;
+	menu->parent = parent;
 	menu->server = server;
 	menu->is_pipemenu_child = waiting_for_pipe_menu;
 	return menu;
@@ -136,11 +131,12 @@ validate(struct server *server)
 }
 
 static struct menuitem *
-item_create(struct menu *menu, const char *text, bool show_arrow)
+item_create(struct menu *menu, const char *text, const char *icon_name, bool show_arrow)
 {
 	assert(menu);
 	assert(text);
 
+	struct theme *theme = menu->server->theme;
 	struct menuitem *menuitem = znew(*menuitem);
 	menuitem->parent = menu;
 	menuitem->selectable = true;
@@ -148,9 +144,17 @@ item_create(struct menu *menu, const char *text, bool show_arrow)
 	menuitem->text = xstrdup(text);
 	menuitem->arrow = show_arrow ? "›" : NULL;
 
+#if HAVE_LIBSFDO
+	if (rc.menu_show_icons && !string_null_or_empty(icon_name)) {
+		menuitem->icon_name = xstrdup(icon_name);
+		menu->has_icons = true;
+	}
+#endif
+
 	menuitem->native_width = font_width(&rc.font_menuitem, text);
 	if (menuitem->arrow) {
-		menuitem->native_width += font_width(&rc.font_menuitem, menuitem->arrow);
+		menuitem->native_width += font_width(&rc.font_menuitem, menuitem->arrow)
+					+ theme->menu_items_padding_x;
 	}
 
 	wl_list_append(&menu->menuitems, &menuitem->link);
@@ -176,7 +180,7 @@ item_create_scene_for_state(struct menuitem *item, float *text_color,
 
 	int bg_width = menu->size.width - 2 * theme->menu_border_width;
 	int arrow_width = item->arrow ?
-		font_width(&rc.font_menuitem, item->arrow) : 0;
+		font_width(&rc.font_menuitem, item->arrow) + theme->menu_items_padding_x : 0;
 	int label_max_width = bg_width - 2 * theme->menu_items_padding_x
 		- arrow_width - icon_width;
 
@@ -226,7 +230,7 @@ item_create_scene_for_state(struct menuitem *item, float *text_color,
 	scaled_font_buffer_update(arrow_buffer, item->arrow, -1,
 		&rc.font_menuitem, text_color, bg_color);
 	/* Vertically center and right-align arrow */
-	x += label_max_width;
+	x += label_max_width + theme->menu_items_padding_x;
 	y = (theme->menu_item_height - label_buffer->height) / 2;
 	wlr_scene_node_set_position(&arrow_buffer->scene_buffer->node, x, y);
 
@@ -243,8 +247,8 @@ item_create_scene(struct menuitem *menuitem, int *item_y)
 
 	/* Menu item root node */
 	menuitem->tree = wlr_scene_tree_create(menu->scene_tree);
-	node_descriptor_create(&menuitem->tree->node,
-		LAB_NODE_DESC_MENUITEM, menuitem);
+	node_descriptor_create(&menuitem->tree->node, LAB_NODE_MENUITEM,
+		/*view*/ NULL, menuitem);
 
 	/* Create scenes for unselected/selected states */
 	menuitem->normal_tree = item_create_scene_for_state(menuitem,
@@ -265,6 +269,8 @@ item_create_scene(struct menuitem *menuitem, int *item_y)
 static struct menuitem *
 separator_create(struct menu *menu, const char *label)
 {
+	assert(menu);
+
 	struct menuitem *menuitem = znew(*menuitem);
 	menuitem->parent = menu;
 	menuitem->selectable = false;
@@ -290,8 +296,8 @@ separator_create_scene(struct menuitem *menuitem, int *item_y)
 
 	/* Menu item root node */
 	menuitem->tree = wlr_scene_tree_create(menu->scene_tree);
-	node_descriptor_create(&menuitem->tree->node,
-		LAB_NODE_DESC_MENUITEM, menuitem);
+	node_descriptor_create(&menuitem->tree->node, LAB_NODE_MENUITEM,
+		/*view*/ NULL, menuitem);
 
 	/* Tree to hold background and line buffer */
 	menuitem->normal_tree = wlr_scene_tree_create(menuitem->tree);
@@ -338,8 +344,8 @@ title_create_scene(struct menuitem *menuitem, int *item_y)
 
 	/* Menu item root node */
 	menuitem->tree = wlr_scene_tree_create(menu->scene_tree);
-	node_descriptor_create(&menuitem->tree->node,
-		LAB_NODE_DESC_MENUITEM, menuitem);
+	node_descriptor_create(&menuitem->tree->node, LAB_NODE_MENUITEM,
+		/*view*/ NULL, menuitem);
 
 	/* Tree to hold background and text buffer */
 	menuitem->normal_tree = wlr_scene_tree_create(menuitem->tree);
@@ -467,42 +473,22 @@ menu_create_scene(struct menu *menu)
  * </item>
  */
 static void
-fill_item(char *nodename, char *content)
+fill_item(struct menu *menu, xmlNode *node)
 {
-	/*
-	 * Nodenames for most menu-items end with '.item.menu' but top-level
-	 * pipemenu items do not have the associated <menu> element so merely
-	 * end with a '.item'
-	 */
-	string_truncate_at_pattern(nodename, ".item.menu");
-	string_truncate_at_pattern(nodename, ".item");
-
-	/* <item label=""> defines the start of a new item */
-	if (!strcmp(nodename, "label")) {
-		current_item = item_create(current_menu, content, false);
-		current_item_action = NULL;
-	} else if (!current_item) {
-		wlr_log(WLR_ERROR, "expect <item label=\"\"> element first. "
-			"nodename: '%s' content: '%s'", nodename, content);
-	} else if (!strcmp(nodename, "icon")) {
-#if HAVE_LIBSFDO
-		if (rc.menu_show_icons && !string_null_or_empty(content)) {
-			xstrdup_replace(current_item->icon_name, content);
-			current_menu->has_icons = true;
-		}
-#endif
-	} else if (!strcmp(nodename, "name.action")) {
-		current_item_action = action_create(content);
-		if (current_item_action) {
-			wl_list_append(&current_item->actions,
-				&current_item_action->link);
-		}
-	} else if (!current_item_action) {
-		wlr_log(WLR_ERROR, "expect <action name=\"\"> element first. "
-			"nodename: '%s' content: '%s'", nodename, content);
-	} else {
-		action_arg_from_xml_node(current_item_action, nodename, content);
+	char *label = (char *)xmlGetProp(node, (xmlChar *)"label");
+	char *icon_name = (char *)xmlGetProp(node, (xmlChar *)"icon");
+	if (!label) {
+		wlr_log(WLR_ERROR, "missing label in <item>");
+		goto out;
 	}
+
+	struct menuitem *item = item_create(menu, label, icon_name, false);
+	lab_xml_expand_dotted_attributes(node);
+	append_parsed_actions(node, &item->actions);
+
+out:
+	xmlFree(label);
+	xmlFree(icon_name);
 }
 
 static void
@@ -518,95 +504,10 @@ item_destroy(struct menuitem *item)
 	free(item);
 }
 
-/*
- * We support XML CDATA for <command> in menu.xml in order to provide backward
- * compatibility with obmenu-generator. For example:
- *
- * <menu id="" label="">
- *   <item label="">
- *     <action name="Execute">
- *       <command><![CDATA[xdg-open .]]></command>
- *     </action>
- *   </item>
- * </menu>
- *
- * <execute> is an old, deprecated openbox variety of <command>. We support it
- * for backward compatibility with old openbox-menu generators. It has the same
- * function and <command>
- *
- * The following nodenames support CDATA.
- *  - command.action.item.*menu.openbox_menu
- *  - execute.action.item.*menu.openbox_menu
- *  - command.action.item.openbox_pipe_menu
- *  - execute.action.item.openbox_pipe_menu
- *  - command.action.item.*menu.openbox_pipe_menu
- *  - execute.action.item.*menu.openbox_pipe_menu
- *
- * The *menu allows nested menus with nodenames such as ...menu.menu... or
- * ...menu.menu.menu... and so on. We could use match_glob() for all of the
- * above but it seems simpler to just check the first three fields.
- */
-static bool
-nodename_supports_cdata(char *nodename)
-{
-	return !strncmp("command.action.", nodename, 15)
-		|| !strncmp("execute.action.", nodename, 15);
-}
-
-static void
-entry(xmlNode *node, char *nodename, char *content)
-{
-	if (!nodename) {
-		return;
-	}
-	xmlChar *cdata = NULL;
-	if (!content && nodename_supports_cdata(nodename)) {
-		cdata = xmlNodeGetContent(node);
-	}
-	if (!content && !cdata) {
-		return;
-	}
-	string_truncate_at_pattern(nodename, ".openbox_menu");
-	string_truncate_at_pattern(nodename, ".openbox_pipe_menu");
-	if (getenv("LABWC_DEBUG_MENU_NODENAMES")) {
-		printf("%s: %s\n", nodename, content ? content : (char *)cdata);
-	}
-	if (in_item) {
-		fill_item(nodename, content ? content : (char *)cdata);
-	}
-	xmlFree(cdata);
-}
-
-static void
-process_node(xmlNode *node)
-{
-	static char buffer[256];
-
-	char *content = (char *)node->content;
-	if (xmlIsBlankNode(node)) {
-		return;
-	}
-	char *name = nodename(node, buffer, sizeof(buffer));
-	entry(node, name, content);
-}
-
-static void xml_tree_walk(xmlNode *node, struct server *server);
-
-static void
-traverse(xmlNode *n, struct server *server)
-{
-	xmlAttr *attr;
-
-	process_node(n);
-	for (attr = n->properties; attr; attr = attr->next) {
-		xml_tree_walk(attr->children, server);
-	}
-	xml_tree_walk(n->children, server);
-}
-
-static bool parse_buf(struct server *server, struct buf *buf);
+static bool parse_buf(struct server *server, struct menu *menu, struct buf *buf);
 static int handle_pipemenu_readable(int fd, uint32_t mask, void *_ctx);
 static int handle_pipemenu_timeout(void *_ctx);
+static void fill_menu_children(struct server *server, struct menu *parent, xmlNode *n);
 
 /*
  * <menu> elements have three different roles:
@@ -615,7 +516,7 @@ static int handle_pipemenu_timeout(void *_ctx);
  *  * Menuitem of submenu type - has ID only
  */
 static void
-handle_menu_element(xmlNode *n, struct server *server)
+fill_menu(struct server *server, struct menu *parent, xmlNode *n)
 {
 	char *label = (char *)xmlGetProp(n, (const xmlChar *)"label");
 	char *icon_name = (char *)xmlGetProp(n, (const xmlChar *)"icon");
@@ -630,9 +531,9 @@ handle_menu_element(xmlNode *n, struct server *server)
 	if (execute && label) {
 		wlr_log(WLR_DEBUG, "pipemenu '%s:%s:%s'", id, label, execute);
 
-		struct menu *pipemenu = menu_create(server, id, label);
+		struct menu *pipemenu = menu_create(server, parent, id, label);
 		pipemenu->execute = xstrdup(execute);
-		if (!current_menu) {
+		if (!parent) {
 			/*
 			 * A pipemenu may not have its parent like:
 			 *
@@ -642,18 +543,16 @@ handle_menu_element(xmlNode *n, struct server *server)
 			 * </openbox_menu>
 			 */
 		} else {
-			current_item = item_create(current_menu, label,
-				/* arrow */ true);
-			fill_item("icon", icon_name);
-			current_item_action = NULL;
-			current_item->submenu = pipemenu;
+			struct menuitem *item = item_create(parent, label,
+				icon_name, /* arrow */ true);
+			item->submenu = pipemenu;
 		}
-	} else if ((label && current_menu) || !current_menu) {
+	} else if ((label && parent) || !parent) {
 		/*
-		 * (label && current_menu) refers to <menu id="" label="">
+		 * (label && parent) refers to <menu id="" label="">
 		 * which is an nested (inline) menu definition.
 		 *
-		 * (!current_menu) catches:
+		 * (!parent) catches:
 		 *     <openbox_menu>
 		 *       <menu id=""></menu>
 		 *     </openbox_menu>
@@ -669,22 +568,20 @@ handle_menu_element(xmlNode *n, struct server *server)
 		 * attribute to make it easier for users to define "root-menu"
 		 * and "client-menu".
 		 */
-		struct menu *parent_menu = current_menu;
-		current_menu = menu_create(server, id, label);
+		struct menu *menu = menu_create(server, parent, id, label);
 		if (icon_name) {
-			current_menu->icon_name = xstrdup(icon_name);
+			menu->icon_name = xstrdup(icon_name);
 		}
-		if (label && parent_menu) {
+		if (label && parent) {
 			/*
 			 * In a nested (inline) menu definition we need to
 			 * create an item pointing to the new submenu
 			 */
-			current_item = item_create(parent_menu, label, true);
-			fill_item("icon", icon_name);
-			current_item->submenu = current_menu;
+			struct menuitem *item = item_create(parent, label,
+				icon_name, true);
+			item->submenu = menu;
 		}
-		traverse(n, server);
-		current_menu = parent_menu;
+		fill_menu_children(server, menu, n);
 	} else {
 		/*
 		 * <menu id=""> (when inside another <menu> element) creates an
@@ -707,7 +604,7 @@ handle_menu_element(xmlNode *n, struct server *server)
 			goto error;
 		}
 
-		struct menu *iter = current_menu;
+		struct menu *iter = parent;
 		while (iter) {
 			if (iter == menu) {
 				wlr_log(WLR_ERROR, "menus with the same id '%s' "
@@ -717,58 +614,55 @@ handle_menu_element(xmlNode *n, struct server *server)
 			iter = iter->parent;
 		}
 
-		current_item = item_create(current_menu, menu->label, true);
-		fill_item("icon", menu->icon_name);
-		current_item->submenu = menu;
+		struct menuitem *item = item_create(parent, menu->label,
+			icon_name ? icon_name : menu->icon_name, true);
+		item->submenu = menu;
 	}
 error:
-	free(label);
-	free(icon_name);
-	free(execute);
-	free(id);
+	xmlFree(label);
+	xmlFree(icon_name);
+	xmlFree(execute);
+	xmlFree(id);
 }
 
 /* This can be one of <separator> and <separator label=""> */
 static void
-handle_separator_element(xmlNode *n)
+fill_separator(struct menu *menu, xmlNode *n)
 {
 	char *label = (char *)xmlGetProp(n, (const xmlChar *)"label");
-	current_item = separator_create(current_menu, label);
-	free(label);
+	separator_create(menu, label);
+	xmlFree(label);
 }
 
+/* parent==NULL when processing toplevel menus in menu.xml */
 static void
-xml_tree_walk(xmlNode *node, struct server *server)
+fill_menu_children(struct server *server, struct menu *parent, xmlNode *n)
 {
-	for (xmlNode *n = node; n && n->name; n = n->next) {
-		if (!strcasecmp((char *)n->name, "comment")) {
-			continue;
-		}
-		if (!strcasecmp((char *)n->name, "menu")) {
-			handle_menu_element(n, server);
-			continue;
-		}
-		if (!strcasecmp((char *)n->name, "separator")) {
-			handle_separator_element(n);
-			continue;
-		}
-		if (!strcasecmp((char *)n->name, "item")) {
-			if (!current_menu) {
+	xmlNode *child;
+	char *key, *content;
+	LAB_XML_FOR_EACH(n, child, key, content) {
+		if (!strcasecmp(key, "menu")) {
+			fill_menu(server, parent, child);
+		} else if (!strcasecmp(key, "separator")) {
+			if (!parent) {
+				wlr_log(WLR_ERROR,
+					"ignoring <separator> without parent <menu>");
+				continue;
+			}
+			fill_separator(parent, child);
+		} else if (!strcasecmp(key, "item")) {
+			if (!parent) {
 				wlr_log(WLR_ERROR,
 					"ignoring <item> without parent <menu>");
 				continue;
 			}
-			in_item = true;
-			traverse(n, server);
-			in_item = false;
-			continue;
+			fill_item(parent, child);
 		}
-		traverse(n, server);
 	}
 }
 
 static bool
-parse_buf(struct server *server, struct buf *buf)
+parse_buf(struct server *server, struct menu *parent, struct buf *buf)
 {
 	int options = 0;
 	xmlDoc *d = xmlReadMemory(buf->data, buf->len, NULL, NULL, options);
@@ -776,36 +670,15 @@ parse_buf(struct server *server, struct buf *buf)
 		wlr_log(WLR_ERROR, "xmlParseMemory()");
 		return false;
 	}
-	xml_tree_walk(xmlDocGetRootElement(d), server);
+
+	xmlNode *root = xmlDocGetRootElement(d);
+	fill_menu_children(server, parent, root);
+
 	xmlFreeDoc(d);
 	xmlCleanupParser();
 	return true;
 }
 
-/*
- * @stream can come from either of the following:
- *   - fopen() in the case of reading a file such as menu.xml
- *   - popen() when processing pipemenus
- */
-static void
-parse_stream(struct server *server, FILE *stream)
-{
-	char *line = NULL;
-	size_t len = 0;
-	struct buf b = BUF_INIT;
-
-	while (getline(&line, &len, stream) != -1) {
-		char *p = strrchr(line, '\n');
-		if (p) {
-			*p = '\0';
-		}
-		buf_add(&b, line);
-	}
-	free(line);
-	parse_buf(server, &b);
-	buf_reset(&b);
-}
-
 static void
 parse_xml(const char *filename, struct server *server)
 {
@@ -818,13 +691,13 @@ parse_xml(const char *filename, struct server *server)
 
 	for (struct wl_list *elm = iter(&paths); elm != &paths; elm = iter(elm)) {
 		struct path *path = wl_container_of(elm, path, link);
-		FILE *stream = fopen(path->string, "r");
-		if (!stream) {
+		struct buf buf = buf_from_file(path->string);
+		if (!buf.len) {
 			continue;
 		}
 		wlr_log(WLR_INFO, "read menu file %s", path->string);
-		parse_stream(server, stream);
-		fclose(stream);
+		parse_buf(server, /*parent*/ NULL, &buf);
+		buf_reset(&buf);
 		if (!should_merge_config) {
 			break;
 		}
@@ -919,11 +792,12 @@ menu_hide_submenu(struct server *server, const char *id)
 	}
 }
 
-static void
-init_client_send_to_menu(struct server *server)
+static struct action *
+item_add_action(struct menuitem *item, const char *action_name)
 {
-	/* Just create placeholder. Contents will be created when launched */
-	menu_create(server, "client-send-to-menu", "");
+	struct action *action = action_create(action_name);
+	wl_list_append(&item->actions, &action->link);
+	return action;
 }
 
 /*
@@ -944,29 +818,35 @@ update_client_send_to_menu(struct server *server)
 
 	struct workspace *workspace;
 
+	/*
+	 * <action name="SendToDesktop"><follow> is true by default so
+	 * GoToDesktop will be called as part of the action.
+	 */
+	struct buf buf = BUF_INIT;
 	wl_list_for_each(workspace, &server->workspaces.all, link) {
 		if (workspace == server->workspaces.current) {
-			char *label = strdup_printf(">%s<", workspace->name);
-			current_item = item_create(menu, label,
-				/*show arrow*/ false);
-			free(label);
+			buf_add_fmt(&buf, ">%s<", workspace->name);
 		} else {
-			current_item = item_create(menu, workspace->name, /*show arrow*/ false);
+			buf_add(&buf, workspace->name);
 		}
-		fill_item("name.action", "SendToDesktop");
-		fill_item("to.action", workspace->name);
+		struct menuitem *item = item_create(menu, buf.data,
+			NULL, /*show arrow*/ false);
+
+		struct action *action = item_add_action(item, "SendToDesktop");
+		action_arg_add_str(action, "to", workspace->name);
+
+		buf_clear(&buf);
 	}
+	buf_reset(&buf);
+
+	separator_create(menu, "");
+	struct menuitem *item = item_create(menu,
+		_("Always on Visible Workspace"), NULL, false);
+	item_add_action(item, "ToggleOmnipresent");
 
 	menu_create_scene(menu);
 }
 
-static void
-init_client_list_combined_menu(struct server *server)
-{
-	/* Just create placeholder. Contents will be created when launched */
-	menu_create(server, "client-list-combined-menu", "");
-}
-
 /*
  * This is client-list-combined-menu an internal menu similar to root-menu and
  * client-menu.
@@ -983,6 +863,7 @@ update_client_list_combined_menu(struct server *server)
 
 	reset_menu(menu);
 
+	struct menuitem *item;
 	struct workspace *workspace;
 	struct view *view;
 	struct buf buffer = BUF_INIT;
@@ -990,32 +871,37 @@ update_client_list_combined_menu(struct server *server)
 	wl_list_for_each(workspace, &server->workspaces.all, link) {
 		buf_add_fmt(&buffer, workspace == server->workspaces.current ? ">%s<" : "%s",
 				workspace->name);
-		current_item = separator_create(menu, buffer.data);
+		separator_create(menu, buffer.data);
 		buf_clear(&buffer);
 
 		wl_list_for_each(view, &server->views, link) {
 			if (view->workspace == workspace) {
-				const char *title = view_get_string_prop(view, "title");
-				if (!view->foreign_toplevel || string_null_or_empty(title)) {
+				if (!view->foreign_toplevel
+						|| string_null_or_empty(view->title)) {
 					continue;
 				}
 
 				if (view == server->active_view) {
 					buf_add(&buffer, "*");
 				}
-				buf_add(&buffer, title);
-
-				current_item = item_create(menu, buffer.data, /*show arrow*/ false);
-				current_item->client_list_view = view;
-				fill_item("name.action", "Focus");
-				fill_item("name.action", "Raise");
+				if (view->minimized) {
+					buf_add_fmt(&buffer, "(%s)", view->title);
+				} else {
+					buf_add(&buffer, view->title);
+				}
+				item = item_create(menu, buffer.data, NULL,
+					/*show arrow*/ false);
+				item->client_list_view = view;
+				item_add_action(item, "Focus");
+				item_add_action(item, "Raise");
 				buf_clear(&buffer);
 				menu->has_icons = true;
 			}
 		}
-		current_item = item_create(menu, _("Go there..."), /*show arrow*/ false);
-		fill_item("name.action", "GoToDesktop");
-		fill_item("to.action", workspace->name);
+		item = item_create(menu, _("Go there..."), NULL,
+			/*show arrow*/ false);
+		struct action *action = item_add_action(item, "GoToDesktop");
+		action_arg_add_str(action, "to", workspace->name);
 	}
 	buf_reset(&buffer);
 	menu_create_scene(menu);
@@ -1025,22 +911,22 @@ static void
 init_rootmenu(struct server *server)
 {
 	struct menu *menu = menu_get_by_id(server, "root-menu");
+	struct menuitem *item;
 
 	/* Default menu if no menu.xml found */
 	if (!menu) {
-		current_menu = NULL;
-		menu = menu_create(server, "root-menu", "");
+		menu = menu_create(server, NULL, "root-menu", "");
 
-		current_item = item_create(menu, _("Terminal"), false);
-		fill_item("name.action", "Execute");
-		fill_item("command.action", "lab-sensible-terminal");
+		item = item_create(menu, _("Terminal"), NULL, false);
+		struct action *action = item_add_action(item, "Execute");
+		action_arg_add_str(action, "command", "lab-sensible-terminal");
 
-		current_item = separator_create(menu, NULL);
+		separator_create(menu, NULL);
 
-		current_item = item_create(menu, _("Reconfigure"), false);
-		fill_item("name.action", "Reconfigure");
-		current_item = item_create(menu, _("Exit"), false);
-		fill_item("name.action", "Exit");
+		item = item_create(menu, _("Reconfigure"), NULL, false);
+		item_add_action(item, "Reconfigure");
+		item = item_create(menu, _("Exit"), NULL, false);
+		item_add_action(item, "Exit");
 	}
 }
 
@@ -1048,46 +934,30 @@ static void
 init_windowmenu(struct server *server)
 {
 	struct menu *menu = menu_get_by_id(server, "client-menu");
+	struct menuitem *item;
 
 	/* Default menu if no menu.xml found */
 	if (!menu) {
-		current_menu = NULL;
-		menu = menu_create(server, "client-menu", "");
-		current_item = item_create(menu, _("Minimize"), false);
-		fill_item("name.action", "Iconify");
-		current_item = item_create(menu, _("Maximize"), false);
-		fill_item("name.action", "ToggleMaximize");
-		current_item = item_create(menu, _("Fullscreen"), false);
-		fill_item("name.action", "ToggleFullscreen");
-		current_item = item_create(menu, _("Roll Up/Down"), false);
-		fill_item("name.action", "ToggleShade");
-		current_item = item_create(menu, _("Decorations"), false);
-		fill_item("name.action", "ToggleDecorations");
-		current_item = item_create(menu, _("Always on Top"), false);
-		fill_item("name.action", "ToggleAlwaysOnTop");
+		menu = menu_create(server, NULL, "client-menu", "");
+		item = item_create(menu, _("Minimize"), NULL, false);
+		item_add_action(item, "Iconify");
+		item = item_create(menu, _("Maximize"), NULL, false);
+		item_add_action(item, "ToggleMaximize");
+		item = item_create(menu, _("Fullscreen"), NULL, false);
+		item_add_action(item, "ToggleFullscreen");
+		item = item_create(menu, _("Roll Up/Down"), NULL, false);
+		item_add_action(item, "ToggleShade");
+		item = item_create(menu, _("Decorations"), NULL, false);
+		item_add_action(item, "ToggleDecorations");
+		item = item_create(menu, _("Always on Top"), NULL, false);
+		item_add_action(item, "ToggleAlwaysOnTop");
 
 		/* Workspace sub-menu */
-		struct menu *workspace_menu = menu_create(server, "workspaces", "");
-		current_item = item_create(workspace_menu, _("Move Left"), false);
-		/*
-		 * <action name="SendToDesktop"><follow> is true by default so
-		 * GoToDesktop will be called as part of the action.
-		 */
-		fill_item("name.action", "SendToDesktop");
-		fill_item("to.action", "left");
-		current_item = item_create(workspace_menu, _("Move Right"), false);
-		fill_item("name.action", "SendToDesktop");
-		fill_item("to.action", "right");
-		current_item = separator_create(workspace_menu, "");
-		current_item = item_create(workspace_menu,
-			_("Always on Visible Workspace"), false);
-		fill_item("name.action", "ToggleOmnipresent");
+		item = item_create(menu, _("Workspace"), NULL, true);
+		item->submenu = menu_get_by_id(server, "client-send-to-menu");
 
-		current_item = item_create(menu, _("Workspace"), true);
-		current_item->submenu = workspace_menu;
-
-		current_item = item_create(menu, _("Close"), false);
-		fill_item("name.action", "Close");
+		item = item_create(menu, _("Close"), NULL, false);
+		item_add_action(item, "Close");
 	}
 
 	if (wl_list_length(&rc.workspace_config.workspaces) == 1) {
@@ -1099,11 +969,14 @@ void
 menu_init(struct server *server)
 {
 	wl_list_init(&server->menus);
+
+	/* Just create placeholder. Contents will be created when launched */
+	menu_create(server, NULL, "client-list-combined-menu", _("Windows"));
+	menu_create(server, NULL, "client-send-to-menu", _("Workspace"));
+
 	parse_xml("menu.xml", server);
 	init_rootmenu(server);
 	init_windowmenu(server);
-	init_client_list_combined_menu(server);
-	init_client_send_to_menu(server);
 	validate(server);
 }
 
@@ -1127,6 +1000,10 @@ nullify_item_pointing_to_this_menu(struct menu *menu)
 		if (iter->parent == menu) {
 			iter->parent = NULL;
 		}
+
+		if (iter->selection.menu == menu) {
+			iter->selection.menu = NULL;
+		}
 	}
 }
 
@@ -1174,11 +1051,6 @@ menu_finish(struct server *server)
 	wl_list_for_each_safe(menu, tmp_menu, &server->menus, link) {
 		menu_free(menu);
 	}
-
-	/* Reset state vars for starting fresh when Reload is triggered */
-	current_item = NULL;
-	current_item_action = NULL;
-	current_menu = NULL;
 }
 
 void
@@ -1334,19 +1206,14 @@ static void
 create_pipe_menu(struct menu_pipe_context *ctx)
 {
 	struct server *server = ctx->pipemenu->server;
-	struct menu *old_current_menu = current_menu;
-	current_menu = ctx->pipemenu;
-	if (!parse_buf(server, &ctx->buf)) {
-		goto restore_menus;
+	if (!parse_buf(server, ctx->pipemenu, &ctx->buf)) {
+		return;
 	}
 	/* TODO: apply validate() only for generated pipemenus */
 	validate(server);
 
 	/* Finally open the new submenu tree */
 	open_menu(ctx->pipemenu, ctx->anchor_rect);
-
-restore_menus:
-	current_menu = old_current_menu;
 }
 
 static void
@@ -1654,15 +1521,6 @@ menu_process_cursor_motion(struct wlr_scene_node *node)
 	menu_process_item_selection(item);
 }
 
-bool
-menu_call_actions(struct wlr_scene_node *node)
-{
-	assert(node && node->data);
-	struct menuitem *item = node_menuitem_from_node(node);
-
-	return menu_execute_item(item);
-}
-
 void
 menu_close_root(struct server *server)
 {
