@@ -3,13 +3,23 @@
 #include <stdbool.h>
 #include <strings.h>
 #include <wlr/backend/libinput.h>
+#include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_input_device.h>
 #include <wlr/types/wlr_keyboard.h>
+#include <wlr/types/wlr_keyboard_group.h>
+#include <wlr/types/wlr_layer_shell_v1.h>
 #include <wlr/types/wlr_pointer.h>
+#include <wlr/types/wlr_pointer_constraints_v1.h>
 #include <wlr/types/wlr_touch.h>
+#include <wlr/types/wlr_virtual_keyboard_v1.h>
+#include <wlr/types/wlr_virtual_pointer_v1.h>
+#include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/log.h>
 #include "common/macros.h"
 #include "common/mem.h"
+#include "config/libinput.h"
+#include "config/rcxml.h"
+#include "config/touch.h"
 #include "input/ime.h"
 #include "input/tablet.h"
 #include "input/tablet-pad.h"
@@ -17,6 +27,8 @@
 #include "input/keyboard.h"
 #include "input/key-state.h"
 #include "labwc.h"
+#include "output.h"
+#include "session-lock.h"
 #include "view.h"
 
 static void
@@ -116,6 +128,7 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 		return;
 	}
 	struct input *input = wlr_input_device->data;
+	assert(input);  /* Check the caller remembered to set this! */
 
 	/* Set scroll factor to 1.0 for Wayland/X11 backends or virtual pointers */
 	if (!wlr_input_device_is_libinput(wlr_input_device)) {
@@ -138,100 +151,148 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 	 */
 	assert(dc);
 
+	wlr_log(WLR_INFO, "configuring input device %s (%s)",
+		libinput_device_get_name(libinput_dev),
+		libinput_device_get_sysname(libinput_dev));
+
+	wlr_log(WLR_INFO, "matched category: %s",
+		dc->name ? dc->name : libinput_device_type_name(dc->type));
+
+	libinput_device_config_tap_set_enabled(libinput_dev,
+		libinput_device_config_tap_get_default_enabled(libinput_dev));
+	libinput_device_config_tap_set_button_map(libinput_dev,
+		libinput_device_config_tap_get_default_button_map(libinput_dev));
 	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0) {
 		wlr_log(WLR_INFO, "tap unavailable");
 	} else {
-		wlr_log(WLR_INFO, "tap configured");
+		wlr_log(WLR_INFO, "tap configured (tap=%d, button_map=%d)",
+			dc->tap, dc->tap_button_map);
 		libinput_device_config_tap_set_enabled(libinput_dev, dc->tap);
 		libinput_device_config_tap_set_button_map(libinput_dev,
 			dc->tap_button_map);
 	}
 
+	libinput_device_config_tap_set_drag_enabled(libinput_dev,
+		libinput_device_config_tap_get_default_drag_enabled(libinput_dev));
 	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0
 			|| dc->tap_and_drag < 0) {
 		wlr_log(WLR_INFO, "tap-and-drag not configured");
 	} else {
-		wlr_log(WLR_INFO, "tap-and-drag configured");
+		wlr_log(WLR_INFO, "tap-and-drag configured (%d)",
+			dc->tap_and_drag);
 		libinput_device_config_tap_set_drag_enabled(
 			libinput_dev, dc->tap_and_drag);
 	}
 
+	libinput_device_config_tap_set_drag_lock_enabled(libinput_dev,
+		libinput_device_config_tap_get_default_drag_lock_enabled(libinput_dev));
 	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0
 			|| dc->drag_lock < 0) {
 		wlr_log(WLR_INFO, "drag lock not configured");
 	} else {
-		wlr_log(WLR_INFO, "drag lock configured");
+		wlr_log(WLR_INFO, "drag lock configured (%d)", dc->drag_lock);
 		libinput_device_config_tap_set_drag_lock_enabled(
 			libinput_dev, dc->drag_lock);
 	}
 
 #if HAVE_LIBINPUT_CONFIG_3FG_DRAG_ENABLED_3FG
+	libinput_device_config_3fg_drag_set_enabled(libinput_dev,
+		libinput_device_config_3fg_drag_get_default_enabled(libinput_dev));
 	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0
 			|| dc->three_finger_drag < 0) {
 		wlr_log(WLR_INFO, "three-finger drag not configured");
 	} else {
-		wlr_log(WLR_INFO, "three-finger drag configured");
+		wlr_log(WLR_INFO, "three-finger drag configured (%d)",
+			dc->three_finger_drag);
 		libinput_device_config_3fg_drag_set_enabled(
 			libinput_dev, dc->three_finger_drag);
 	}
 #endif
 
+	libinput_device_config_scroll_set_natural_scroll_enabled(libinput_dev,
+		libinput_device_config_scroll_get_default_natural_scroll_enabled(libinput_dev));
 	if (libinput_device_config_scroll_has_natural_scroll(libinput_dev) <= 0
 			|| dc->natural_scroll < 0) {
 		wlr_log(WLR_INFO, "natural scroll not configured");
 	} else {
-		wlr_log(WLR_INFO, "natural scroll configured");
+		wlr_log(WLR_INFO, "natural scroll configured (%d)",
+			dc->natural_scroll);
 		libinput_device_config_scroll_set_natural_scroll_enabled(
 			libinput_dev, dc->natural_scroll);
 	}
 
+	libinput_device_config_left_handed_set(libinput_dev,
+		libinput_device_config_left_handed_get_default(libinput_dev));
 	if (libinput_device_config_left_handed_is_available(libinput_dev) <= 0
 			|| dc->left_handed < 0) {
 		wlr_log(WLR_INFO, "left-handed mode not configured");
 	} else {
-		wlr_log(WLR_INFO, "left-handed mode configured");
+		wlr_log(WLR_INFO, "left-handed mode configured (%d)",
+			dc->left_handed);
 		libinput_device_config_left_handed_set(libinput_dev,
 			dc->left_handed);
 	}
 
+	libinput_device_config_accel_set_speed(libinput_dev,
+		libinput_device_config_accel_get_default_speed(libinput_dev));
+	libinput_device_config_accel_set_profile(libinput_dev,
+		libinput_device_config_accel_get_default_profile(libinput_dev));
 	if (libinput_device_config_accel_is_available(libinput_dev) == 0) {
 		wlr_log(WLR_INFO, "pointer acceleration unavailable");
 	} else {
-		wlr_log(WLR_INFO, "pointer acceleration configured");
 		if (dc->pointer_speed >= -1) {
+			wlr_log(WLR_INFO, "pointer speed configured (%g)",
+				dc->pointer_speed);
 			libinput_device_config_accel_set_speed(libinput_dev,
 				dc->pointer_speed);
+		} else {
+			wlr_log(WLR_INFO, "pointer speed not configured");
 		}
+
 		if (dc->accel_profile > 0) {
+			wlr_log(WLR_INFO,
+				"pointer accel profile configured (%d)",
+				dc->accel_profile);
 			libinput_device_config_accel_set_profile(libinput_dev,
 				dc->accel_profile);
+		} else {
+			wlr_log(WLR_INFO,
+				"pointer accel profile not configured");
 		}
 	}
 
+	libinput_device_config_middle_emulation_set_enabled(libinput_dev,
+		libinput_device_config_middle_emulation_get_default_enabled(libinput_dev));
 	if (libinput_device_config_middle_emulation_is_available(libinput_dev)
 			== 0 || dc->middle_emu < 0)  {
 		wlr_log(WLR_INFO, "middle emulation not configured");
 	} else {
-		wlr_log(WLR_INFO, "middle emulation configured");
+		wlr_log(WLR_INFO, "middle emulation configured (%d)",
+			dc->middle_emu);
 		libinput_device_config_middle_emulation_set_enabled(
 			libinput_dev, dc->middle_emu);
 	}
 
+	libinput_device_config_dwt_set_enabled(libinput_dev,
+		libinput_device_config_dwt_get_default_enabled(libinput_dev));
 	if (libinput_device_config_dwt_is_available(libinput_dev) == 0
 			|| dc->dwt < 0) {
 		wlr_log(WLR_INFO, "dwt not configured");
 	} else {
-		wlr_log(WLR_INFO, "dwt configured");
+		wlr_log(WLR_INFO, "dwt configured (%d)", dc->dwt);
 		libinput_device_config_dwt_set_enabled(libinput_dev, dc->dwt);
 	}
 
+	libinput_device_config_click_set_method(libinput_dev,
+		libinput_device_config_click_get_default_method(libinput_dev));
 	if ((dc->click_method != LIBINPUT_CONFIG_CLICK_METHOD_NONE
 			&& (libinput_device_config_click_get_methods(libinput_dev)
 				& dc->click_method) == 0)
 			|| dc->click_method < 0) {
 		wlr_log(WLR_INFO, "click method not configured");
 	} else {
-		wlr_log(WLR_INFO, "click method configured");
+		wlr_log(WLR_INFO, "click method configured (%d)",
+			dc->click_method);
 
 		/*
 		 * Note, the documentation claims that:
@@ -245,6 +306,8 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 		libinput_device_config_click_set_method(libinput_dev, dc->click_method);
 	}
 
+	libinput_device_config_scroll_set_method(libinput_dev,
+		libinput_device_config_scroll_get_default_method(libinput_dev));
 	if (dc->scroll_method < 0) {
 		wlr_log(WLR_INFO, "scroll method not configured");
 	} else if (dc->scroll_method != LIBINPUT_CONFIG_SCROLL_NO_SCROLL
@@ -252,20 +315,27 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 				& dc->scroll_method) == 0) {
 		wlr_log(WLR_INFO, "scroll method not supported");
 	} else {
-		wlr_log(WLR_INFO, "scroll method configured");
+		wlr_log(WLR_INFO, "scroll method configured (%d)",
+			dc->scroll_method);
 		libinput_device_config_scroll_set_method(libinput_dev, dc->scroll_method);
 	}
 
+	libinput_device_config_send_events_set_mode(libinput_dev,
+		libinput_device_config_send_events_get_default_mode(libinput_dev));
 	if ((dc->send_events_mode != LIBINPUT_CONFIG_SEND_EVENTS_ENABLED
 			&& (libinput_device_config_send_events_get_modes(libinput_dev)
 				& dc->send_events_mode) == 0)
 			|| dc->send_events_mode < 0) {
 		wlr_log(WLR_INFO, "send events mode not configured");
 	} else {
-		wlr_log(WLR_INFO, "send events mode configured");
+		wlr_log(WLR_INFO, "send events mode configured (%d)",
+			dc->send_events_mode);
 		libinput_device_config_send_events_set_mode(libinput_dev, dc->send_events_mode);
 	}
 
+	float default_matrix[6];
+	libinput_device_config_calibration_get_default_matrix(libinput_dev, default_matrix);
+	libinput_device_config_calibration_set_matrix(libinput_dev, default_matrix);
 	/* Non-zero if the device can be calibrated, zero otherwise. */
 	if (libinput_device_config_calibration_has_matrix(libinput_dev) == 0
 			|| !dc->have_calibration_matrix) {
@@ -275,7 +345,7 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 		libinput_device_config_calibration_set_matrix(libinput_dev, dc->calibration_matrix);
 	}
 
-	wlr_log(WLR_INFO, "scroll factor configured");
+	wlr_log(WLR_INFO, "scroll factor configured (%g)", dc->scroll_factor);
 	input->scroll_factor = dc->scroll_factor;
 }
 
@@ -336,6 +406,9 @@ new_keyboard(struct seat *seat, struct wlr_input_device *device, bool is_virtual
 	keyboard->base.wlr_input_device = device;
 	keyboard->wlr_keyboard = kb;
 	keyboard->is_virtual = is_virtual;
+	device->data = keyboard;
+
+	configure_libinput(device);
 
 	if (!seat->keyboard_group->keyboard.keymap) {
 		wlr_log(WLR_ERROR, "cannot set keymap");
@@ -488,9 +561,9 @@ handle_new_input(struct wl_listener *listener, void *data)
 }
 
 static void
-new_virtual_pointer(struct wl_listener *listener, void *data)
+handle_new_virtual_pointer(struct wl_listener *listener, void *data)
 {
-	struct seat *seat = wl_container_of(listener, seat, virtual_pointer_new);
+	struct seat *seat = wl_container_of(listener, seat, new_virtual_pointer);
 	struct wlr_virtual_pointer_v1_new_pointer_event *event = data;
 	struct wlr_virtual_pointer_v1 *pointer = event->new_pointer;
 	struct wlr_input_device *device = &pointer->pointer.base;
@@ -574,9 +647,7 @@ seat_init(struct server *server)
 
 	seat->virtual_pointer = wlr_virtual_pointer_manager_v1_create(
 		server->wl_display);
-	wl_signal_add(&seat->virtual_pointer->events.new_virtual_pointer,
-		&seat->virtual_pointer_new);
-	seat->virtual_pointer_new.notify = new_virtual_pointer;
+	CONNECT_SIGNAL(seat->virtual_pointer, seat, new_virtual_pointer);
 
 	seat->virtual_keyboard = wlr_virtual_keyboard_manager_v1_create(
 		server->wl_display);
@@ -606,7 +677,7 @@ seat_finish(struct server *server)
 	struct seat *seat = &server->seat;
 	wl_list_remove(&seat->new_input.link);
 	wl_list_remove(&seat->focus_change.link);
-	wl_list_remove(&seat->virtual_pointer_new.link);
+	wl_list_remove(&seat->new_virtual_pointer.link);
 	wl_list_remove(&seat->new_virtual_keyboard.link);
 
 	struct input *input, *next;
@@ -663,11 +734,12 @@ seat_reconfigure(struct server *server)
 	struct seat *seat = &server->seat;
 	struct input *input;
 	cursor_reload(seat);
-	overlay_reconfigure(seat);
+	overlay_finish(seat);
 	keyboard_reset_current_keybind();
 	wl_list_for_each(input, &seat->inputs, link) {
 		switch (input->wlr_input_device->type) {
 		case WLR_INPUT_DEVICE_KEYBOARD:
+			configure_libinput(input->wlr_input_device);
 			configure_keyboard(seat, input);
 			break;
 		case WLR_INPUT_DEVICE_POINTER:
@@ -776,46 +848,6 @@ seat_set_focus_layer(struct seat *seat, struct wlr_layer_surface_v1 *layer)
 	seat->focused_layer = layer;
 }
 
-static void
-pressed_surface_destroy(struct wl_listener *listener, void *data)
-{
-	struct seat *seat = wl_container_of(listener, seat,
-		pressed_surface_destroy);
-
-	/*
-	 * Using data directly prevents 'unused variable'
-	 * warning when compiling without asserts
-	 */
-	assert(data == seat->pressed.surface);
-
-	seat_reset_pressed(seat);
-}
-
-void
-seat_set_pressed(struct seat *seat, struct cursor_context *ctx)
-{
-	assert(ctx);
-	assert(ctx->view || ctx->surface);
-	seat_reset_pressed(seat);
-
-	seat->pressed = *ctx;
-
-	if (ctx->surface) {
-		seat->pressed_surface_destroy.notify = pressed_surface_destroy;
-		wl_signal_add(&ctx->surface->events.destroy,
-			&seat->pressed_surface_destroy);
-	}
-}
-
-void
-seat_reset_pressed(struct seat *seat)
-{
-	if (seat->pressed.surface) {
-		wl_list_remove(&seat->pressed_surface_destroy.link);
-	}
-	seat->pressed = (struct cursor_context){0};
-}
-
 void
 seat_output_layout_changed(struct seat *seat)
 {
